
#include <Arduino.h>
#include <SPI.h>

#define EMISSION 2 // PE4, à l'état haut pendant l'émission
#define ENA 5 // commande ENABLE du L298
#define SIGNAL1 11 // PB5 (OC1A) commande IN1 du L298
#define SIGNAL2 12 // PB6 (OC1B) commande IN2 du L298
#define ADCSAMPLE 3 // PE5, signal de contrôle de l'échantillonnage
#define RECEPT 18 // INT3, sortie du comparateur (COMP)
#define POT_COMP 10 // CS COMP
#define POT_AMP 13 // CS AMP
#define BUFSIZE 255
#define PERIOD 25 // période du signal de commande en microsecondes
#define NBCYCLES 10 // nombre de cycles
#define RECEPTIME 3000 // délai (en us) entre l'émission et le début de la réception
#define SEUILREL 0.1 // valeur du seuil relativement à la valeur maximale de l'enveloppe
#define VCC_DETECT 5.0 // tension d'alimentation du détecteur d'enveloppe

uint16_t diviseur[6] = {0,1,8,64,256,1024};
uint32_t timer_top;
volatile uint32_t count, nbCount;
float periode_timer;
uint8_t d;
volatile uint8_t adc_buffer[BUFSIZE];
uint16_t buffer_index;
volatile uint8_t adcflag;
volatile uint8_t recept_wait;
volatile float temps_trajet;
uint32_t recept_count_min;
uint8_t potAmpli;
uint32_t max_count;
uint8_t tccr1a,tccr1b;            
             
void config_train_impulsions(uint32_t period, int32_t nombre, float ra, float rb) {
  max_count = nombre+1;
  recept_count_min = RECEPTIME/period;
  cli();
  tccr1a = 0;
  tccr1b = 0;
  //phase and frequency correct PWM mode, top = ICR1 WGM3:0 = 1,0,0,0
  tccr1b |= (1 << WGM13);
  // COM1A1 = 1, COM1A0 = 0 : clear OC1A on compare match when up counting
  // COM1B1 = 1, COM1B0 = 1 : set OC1B on compare match when up counting
  tccr1a |= (1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0);
  uint32_t icr = (F_CPU/1000000*period/2);
  int d = 1;
  while ((icr>0xFFFF)&&(d<5)) {
     d++;
     icr = (F_CPU/1000000*period/2/diviseur[d]);
   }
  timer_top = icr;
  periode_timer = icr*2*diviseur[d]*1000000/F_CPU;
  tccr1b |= d;
  ICR1 = icr;
  OCR1A = icr*ra;
  OCR1B = icr*(1.0-rb);
  TIMSK1 = (1 << TOIE1) ; // overflow interrupt enable
  TCNT1 = 0;
  count = 0;
  sei();
}                
                 
ISR(TIMER1_OVF_vect) {
  count++;
  if (count==max_count) {
    TCCR1A = 0; // arrêt de l'émission
    PORTB &= ~(1 << PORTB5);
    PORTB &= ~(1 << PORTB6);
    PORTE &= ~(1 << PORTE4);
  }
}                 
                 
void start_impulsions() {
  cli();
  TCNT1 = 0;
  count = 0;
  TCCR1A = tccr1a;
  TCCR1B = tccr1b;
  buffer_index = 0;
  recept_wait = 1;
  PORTE |= (1 << PORTE4);
  sei();
}                
                 
void config_adc() {
  ADMUX = 0; // voie 0
  ADCSRB = 0;
  ADMUX |= (1 << REFS0); // Voltage reference : AVCC with external capacitor at AREF pin
  ADMUX |= (1 << ADLAR);
  ADCSRA = (1 << ADEN); // enable ADC
  ADCSRA |= (1 << ADATE); // Auto Trigger Enable
  ADCSRA |= (1 << ADIE); // interrupt enable
  ADCSRB |= (1 << ADTS2)|(1 << ADTS1);// trigger source : timer 1 overflow
  ADCSRA |= 4; // prescaler
  buffer_index = 0;
}                
                 
ISR(ADC_vect) {
  uint8_t x = ADCH;
  if (count < recept_count_min) return;
  if (buffer_index < BUFSIZE) {
    adc_buffer[buffer_index] = x;
    buffer_index += 1;
  }
  if (adcflag==0) {
      adcflag = 1;
      PORTE |= (1 << PORTE5);
  }
  else {
      adcflag = 0;
      PORTE &= ~(1 << PORTE5);
  }
}                   
                    
ISR(INT3_vect) {
  uint32_t tcn = TCNT1;
  if (count < recept_count_min) return;
  if (recept_wait) {
      recept_wait = 0;
      temps_trajet = periode_timer*(count+tcn*0.5/timer_top);
  }
  
}              
                    
void potentiometre_comp(int v) {
  digitalWrite(POT_COMP, LOW);
  delayMicroseconds(500);
  SPI.transfer(0x00);
  SPI.transfer(v);
  digitalWrite(POT_COMP, HIGH);
}

void potentiometre_amp(int v) {
  digitalWrite(POT_AMP, LOW);
  delayMicroseconds(100);
  SPI.transfer(0x00);
  SPI.transfer(v);
  digitalWrite(POT_AMP, HIGH);
}          
                    
void setup() {
      Serial.begin(115200);
      SPI.begin();
      SPI.beginTransaction(SPISettings(10000000, MSBFIRST, SPI_MODE0));
      pinMode(POT1,OUTPUT);
      pinMode(POT2,OUTPUT);
      digitalWrite(POT_COMP,HIGH);
      digitalWrite(POT_AMP,HIGH);
      pinMode(ADCSAMPLE,OUTPUT);
      pinMode(SIGNAL1,OUTPUT);
      pinMode(SIGNAL2,OUTPUT);
      pinMode(EMISSION,OUTPUT);
      PORTB &= ~(1 << PORTB5);
      PORTB &= ~(1 << PORTB6);
      pinMode(ENA,OUTPUT);
      digitalWrite(ENA,HIGH);
      EICRA |= (1<<ISC31); // interruption INT3 (pin 18) programmée sur le front descendant
      EIMSK |= (1<<INT3); // activation de INT3
      adcflag = 0;
      delay(100);
      potentiometre1(110); // potentiomètre de réglage du seuil du comparateur
      potAmpli = 50;
      potentiometre2(potAmpli); // potentiomètre de réglage du gain de l'amplificateur
      config_adc();
      float R=0.5;
      float r2=R-6e-2;
      float r1=1-R-6e-2;
      config_train_impulsions(PERIOD,NBCYCLES,r1,r2);
}                      
                      
void emission_reglage() {
      start_adc();
      start_impulsions();
      delay(100);
      uint8_t envmax = 0;
      for (int i=0; i<BUFSIZE; i++) {
        if (adc_buffer[i] > envmax) envmax = adc_buffer[i];
      }
      Serial.print("Max = "); Serial.println(envmax);
      float Umax = (envmax-128)*5.0/255; // max de l'enveloppe en V, par rapport à 2.5 V
      Serial.print("Umax = "); Serial.println(Umax);
      float seuil = Umax*SEUILREL + 2.5; // tension de seuil du comparateur
      Serial.print("Seuil = "); Serial.println(seuil);
      int p = 127.0*(VCC_DETECT / seuil -1);
      Serial.print("potComp = "); Serial.println(p);
      potentiometre_comp(p);
      if ((Umax > 1.4)&&(potAmpli>0)) {potAmpli -=1; potentiometre_amp(potAmpli);}
      else if ((Umax < 1.4)&&(potAmpli <= 128)) {potAmpli +=1; potentiometre_amp(potAmpli);}
      Serial.print("potAmpli = "); Serial.println(potAmpli);
      Serial.print("temps trajet = "); Serial.println(temps_trajet);
}                                           
                      
void emission_mesure() {
  stop_adc();
  start_impulsions();
  delay(100);
  Serial.print("--temps trajet = "); Serial.println(temps_trajet);
}                      
                       
void loop() {
    for (int k=0; k<5; k++) emission_reglage();
    for (int k=0; k<5; k++) emission_mesure(); 
}                       
                       